/*********************************************
 *             GIF viewer on SUNVIEW         *
 *                                           *
 *      March 23 1989 By Marcel J.E. Mol     *
 *                                           *
 * I hereby place this program               *
 * in the public domain, i.e. there are no   *
 * copying restrictions of any kind.         *
 *********************************************/

/* 
 * Here is a gif viewer for the SUN running suntools (or sunview).
 * It is intended for color monitors (8 planes) and I don't
 * know how it will react on a monochrome screen.
 * It is rather slow because it uses call to sunview for 
 * each pixel. Should be converted to use rasterfiles one day...
 * 
 * Little help:
 *     left button: exit a picture being drawn, and starts the following
 *                  picture, if any.
 *     right button: exit program immediately.
 *     middle button: when a picture is ready pressing the middle button
 *                    will start the following picture, if any.
 * 
 * Usage:
 *        gif2sun [-sn -ln -b] <gif-files>
 * 
 *       -sn : tries to expand a picture n times.
 *       -ln : number of lines to hold for sunview batch mode.
 *       -b  : disable optimization done by not drawing pixels in 
 *             the background color.
 * 
 * Compile:
 *          cc gif2sun.c -o gif2sun -lsuntool -lsunwindow -lpixrect
 * 
 */ 

#define FAST

#include <stdio.h>
#include <suntool/sunview.h>
#include <suntool/panel.h>
#include <suntool/canvas.h>

/* Change event_action to event_id(event) for pre 4.0 */
#ifndef event_action
#define event_action event_id
#define oldSunOS
#endif

#ifndef oldSunOS
#include <suntool/alert.h>	/* Only for SunOS 4 */
#include <alloca.h>		/* Cannot find this on 3.5 */
#endif

char *malloc();
int strncmp();

#define FALSE 0
#define TRUE 1
#define COLSIZE 256
#define PICSIZE 1;
#define SHOWCOUNT 10;

#define BUTTONFONT "/usr/lib/fonts/fixedwidthfonts/gallant.r.19"
static char *default_font =
       "DEFAULT_FONT=/usr/lib/fonts/fixedwidthfonts/screen.r.13";

/*
 * Sunview variables 
 */
static Frame fr;
static Canvas cv;
static Pixwin *pw;
static Panel pn;
static Pixfont *bf, *pf;

/*
 * LZW structures and variables
 */
typedef int bool;
unsigned char *stackp;
unsigned int prefix[4096];
unsigned char suffix[4096];
unsigned char stack[4096];
int datasize,codesize,codemask;     /* Decoder working variables */
int clear,eoi;                      /* Special code values */
int avail;
int oldcode;

/*
 * GIF variables
 */
FILE *infile;
unsigned int screenwidth;           /* The dimensions of the screen */
unsigned int screenheight;          /*   (not those of the image)   */
bool global;                        /* Is there a global color map? */
int globalbits;                     /* Number of bits of global colors */
unsigned char globalmap[COLSIZE][3];/* RGB values for global color map */
char bgcolor;                       /* background color */
unsigned char *raster;              /* Decoded image data */
unsigned left,top,width,height;

char *progname;
char *filename;
unsigned int windowwidth = 0;
unsigned int windowheight = 0;
unsigned int size;                  /* scale factor */

unsigned int initsize = PICSIZE;
int endpicflag = FALSE;
int drawflag;
unsigned int showcount = SHOWCOUNT;
int dobgflag = FALSE;

void convert();
int  checksignature();
void readscreen();
int  readimage();
void readextension();
int  readraster();
int  process();
void outcode();
void initsunview();
void setupsunview();
void initcolors();
int  rasterize();
void waitevent();
void handlevent();
void usage();




main(argc,argv)
int argc;
char *argv[];
{
    extern int optind;
    extern char *optarg;
    int flag;

    progname = *argv;
    while ((flag = getopt(argc, argv, "s:l:b")) != EOF) {
        switch (flag) {
            case 's' : if ((initsize = atoi(optarg)) <= 0) {
                           initsize = PICSIZE;
                           fprintf(stderr, "scale factor must more than 1\n");
                       }
                       break;
            case 'l' : if ((showcount = atoi(optarg)) <= 0) {
                           showcount = SHOWCOUNT;
                           fprintf(stderr, "showcount must more than 1\n");
                       }
                       break;
            case 'b' : dobgflag = TRUE;
                       break;
            default  : fprintf(stderr, "ignoring unknown flag %c\n", flag);
                       usage();
        }
    }

    initsunview();
    if (optind >= argc) {
        filename = "stdin";
        convert();
    }
    else
        while (optind < argc) {
            filename = argv[optind];
            optind++;
            if ((infile = fopen(filename,"r")) == NULL) {
                perror(filename);
                continue;
            }
            convert();
            fclose(infile);
        }

    if (drawflag)        /* wait for last picture if fully drawn */
        waitevent();

} /* main */



void convert()
{
    char ch;

    printf("%s:\n", filename);
    if (checksignature())
        return;
    readscreen();
    size = initsize;
    while ((ch = getc(infile)) != ';' && ch != EOF) {
        switch (ch) {
            case '\0':  break;  /* this kludge for non-standard files */
            case ',':   if (readimage())
                           return;
                        break;
            case '!':   readextension();
                        break;
            default:    printf(stderr, "illegal GIF block type\n");
                        return;
                        break;
        }
    }

} /* convert */



checksignature()
{
    char buf[6];

    fread(buf,1,6,infile);
    if (strncmp(buf,"GIF",3)) {
        fprintf(stderr, "file is not a GIF file\n");
        return 1;
    }
    if (strncmp(&buf[3],"87a",3)) {
        fprintf(stderr, "unknown GIF version number\n");
        return 1;
    }
    return 0;

} /* checksignature */



/*
 * Get information which is global to all the images stored in the file
 */

void readscreen()
{
    unsigned char buf[7];

    fread(buf,1,7,infile);
    screenwidth = buf[0] + (buf[1] << 8);
    screenheight = buf[2] + (buf[3] << 8);
    global = buf[4] & 0x80;
    if (global) {
        globalbits = (buf[4] & 0x07) + 1;
        fread(globalmap,3,1<<globalbits,infile);
    }
    bgcolor = buf[5];
    printf("    global screen: %dx%dx%d, backgroundcolor: %d\n",
                screenwidth, screenheight, 1<<globalbits, bgcolor);

} /* readscreen */




readimage()
{
    unsigned char buf[9];
    bool local, interleaved;
    char localmap[256][3];
    int localbits;
    register row;
    register i;
    static int waitflag = FALSE;

    fread(buf, 1, 9, infile);
    left = buf[0] + (buf[1] << 8);
    top = buf[2] + (buf[3] << 8);
    width = buf[4] + (buf[5] << 8);
    height = buf[6] + (buf[7] << 8);
    local = buf[8] & 0x80;
    interleaved = buf[8] & 0x40;
    printf("    image: %dx%d %s  org: %d,%d\n", width, height,
              interleaved ? "interleaved" : "", left, top);
    if (local == 0 && global == 0) {
        fprintf(stderr, "no colormap present for image\n");
        return 1;
    }
    if ((raster = (unsigned char*) malloc(width*height)) == NULL) {
        fprintf(stderr, "not enough memory for image\n");
        return 1;
    }
    if (readraster(width, height))
        return 1;

    if (waitflag)         /* if first picture or previous picture is   */
        waitevent();      /* interrupted, don't wait for a buttonpress */

    setupsunview(screenwidth, screenheight);
    if (local) {
        localbits = (buf[8] & 0x7) + 1;
        printf("   local colors: %d\n", 1<<localbits);
        fread(localmap, 3, 1<<localbits, infile);
        initcolors(localmap, 1<<localbits, bgcolor);
    } else if (global)
        initcolors(globalmap, 1<<globalbits, bgcolor);

    waitflag = rasterize(interleaved, raster);
    free(raster);

    return 0;

} /* readimage */



/*
 * Read a GIF extension block (and do nothing with it).
 */

void readextension()
{
    unsigned char code;
    int count;
    char buf[255];

    code = getc(infile);
    while (count = getc(infile)) fread(buf, 1, count, infile);

} /* readextension */



/*
 * Decode a raster image
 */

readraster(width, height)
unsigned width,height;
{
        unsigned char *fill = raster;
        unsigned char buf[255];
        register bits=0;
        register unsigned datum=0;
        register unsigned char *ch;
        register int count, code;

        datasize = getc(infile);
        clear = 1 << datasize;
        eoi = clear + 1;
        avail = clear + 2;
        oldcode = -1;
        codesize = datasize + 1;
        codemask = (1 << codesize) - 1;
        for (code = 0; code < clear; code++) {
            prefix[code] = 0;
            suffix[code] = code;
        }
        stackp = stack;
        for (count = getc(infile); count > 0; count = getc(infile)) {
            fread(buf,1,count,infile);
            for (ch=buf; count-- > 0; ch++) {
                datum += *ch << bits;
                bits += 8;
                while (bits >= codesize) {
                    code = datum & codemask;
                    datum >>= codesize;
                    bits -= codesize;
                    if (code == eoi) {               /* This kludge put in */
#if defined(DEBUG)
                        printf("found eoi code\n");
#endif
                        goto exitloop;               /* because some GIF files */
                    }                                /* aren't standard */
                    if (process(code, &fill)) 
                        goto exitloop;
                }
            }
            if (fill >= raster + width*height) {
                fprintf(stderr, "raster full before eoi code\n");
                goto exitloop;
            }
        }
exitloop:
        if (fill != raster + width*height)  {
            fprintf(stderr, "warning: wrong rastersize: %ld bytes\n",
                      (long) (fill-raster));
            fprintf(stderr, "         instead of %ld bytes\n",
                      (long) width*height);
            return 0;  /* but try to show picture ... */
        }

        return 0;

} /* readraster */




/*
 * Process a compression code.  "clear" resets the code table.  Otherwise
 * make a new code table entry, and output the bytes associated with the
 * code.
 */

process(code, fill)
register code;
unsigned char **fill;
{
        int incode;
        static unsigned char firstchar;

        if (code == clear) {
#if defined(DEBUG)
            fprintf(stderr, "clear encountered\n");
#endif
            codesize = datasize + 1;
            codemask = (1 << codesize) - 1;
            avail = clear + 2;
            oldcode = -1;
            return 0;
        }

        if (code > avail) {
            fprintf(stderr, "code %d to large for %d\n", code, avail);
            code = avail;
            return 1;  
        }

        if (oldcode == -1) {
            *(*fill)++ = suffix[code];
            firstchar = oldcode = code;
            return 0;
        }

        incode = code;
        if (code == avail) {      /* the first code is always < avail */
            *stackp++ = firstchar;
            code = oldcode;
        }
        while (code > clear) {
            *stackp++ = suffix[code];
            code = prefix[code];
        }

        *stackp++ = firstchar = suffix[code];
        prefix[avail] = oldcode;
        suffix[avail] = firstchar;
if (avail < 4096)
        avail++;

        if (((avail & codemask) == 0) && (avail < 4096)) {
            codesize++;
            codemask += avail;
        }
#if defined(DEBUG)
        if (avail >= 4096) {
            fprintf(stderr, "tables full, avail = %d\n", avail);
        }
#endif

        oldcode = incode;
        do {
            *(*fill)++ = *--stackp;
        } while (stackp > stack);

        return 0;

} /* process */



void initsunview()
{

   (void) putenv(default_font);
    bf = pf_open(BUTTONFONT);
    pf = pf_default();

} /* initsunview */



void setupsunview(width, height)
unsigned width, height;
{
    static havewin = FALSE;
    char title[128];
    unsigned rwidth;

    rwidth = width + width%2;    /* compensate for odd width */
    while ((size*rwidth > 1000) && (size > 1))
        size--;
    rwidth *= size;
    height *= size;
    if ((windowwidth < rwidth) || (windowheight < height)) {
        windowwidth = rwidth;
        windowheight = height;
/*        if (havewin)    DESTROYING WINDOWS RESULT IN SEGMENTATION FAULTS 
            window_destroy(fr); */
        fr = window_create(0, FRAME,
             WIN_X, 5,
             WIN_Y, 50,
             WIN_SHOW, TRUE,
             FRAME_LABEL, "Gif to Sun",
             0);
        havewin = TRUE;
        pn = window_create(fr, PANEL, 0);
        panel_create_item(pn, PANEL_MESSAGE,
                           PANEL_ITEM_X, 5,
                           PANEL_ITEM_Y, 5,
                           PANEL_LABEL_STRING, "File:",
                           0);
        window_fit_height(pn);
        cv = window_create(fr, CANVAS,
		           WIN_WIDTH, windowwidth,
		           WIN_HEIGHT, windowheight,
		           WIN_CONSUME_PICK_EVENT, LOC_DRAG, 
		           WIN_EVENT_PROC, handlevent, 
		           0);

        window_fit(fr);
        pw = canvas_pixwin(cv);
        notify_interpose_destroy_func(fr, handlevent);
        pw_setcmsname(pw, "gif colors");
    }

    notify_dispatch();     /* show the correct window */
    sprintf(title, "%-20s", filename);
    panel_pw_text(pn, 50, 5+panel_fonthome(pf), PIX_SRC, pf, title);

} /* setupsunview */



/*
 * Convert a color map (local or global) to arrays with R, G and B
 * values. Pass colors to SUNVIEW and set the background color.
 */

void initcolors(colormap, ncolors, bgcolor)
unsigned char colormap[COLSIZE][3];
int ncolors;
int bgcolor;
{
    register i;
    unsigned char red[COLSIZE];
    unsigned char green[COLSIZE];
    unsigned char blue[COLSIZE];

#if defined(DEBUG)
    printf("            ");
#endif
    for (i = 0; i < ncolors; i++) {
        red[i]   = colormap[i][0];
        green[i] = colormap[i][1];
        blue[i]  = colormap[i][2];
#if defined(DEBUG)
        printf("   %3u: %3u, %3u, %3u", i, red[i], green[i], blue[i]);
        if (i%3 == 2)
            printf("\n            ");
#endif
    }
    pw_putcolormap(pw, 0, COLSIZE, red, green, blue);

#if defined(DEBUG)
    printf("\n      Background: %3u: %3u, %3u, %3u\n", bgcolor,
                          red[bgcolor], green[bgcolor], blue[bgcolor]);
#endif
    pw_writebackground(pw, 0, 0, windowwidth, windowheight,
                           PIX_SET | PIX_COLOR(bgcolor));

} /* initcolors */



/*
 * Read a row out of the raster image and write it to the screen
 */
#ifdef FAST
rasterize(interleaved, raster)
int interleaved;
register unsigned char *raster;
{
    register row, col;
    register unsigned char *rr;
    unsigned char * newras;
    Pixrect *picture;
    int rwidth;
    
#define DRAWSEGMENT(offset, step)                             \
        for (row = offset; row < height; row += step) {       \
            rr = newras + row*rwidth;                          \
            bcopy(raster, rr, rwidth);                         \
            raster += width;                                  \
        }


    panel_pw_text(pn, 250, 5+panel_fonthome(pf), PIX_SRC, pf, "Drawing");
    drawflag = TRUE;
    rwidth = width + width%2;    /* compensate for odd width */
        if ((newras = (unsigned char*) malloc(rwidth*height)) == NULL) {
            fprintf(stderr, "not enough memory for image\n");
            return 1;
        }
    if (interleaved) {
        DRAWSEGMENT(0, 8);
        DRAWSEGMENT(4, 8);
        DRAWSEGMENT(2, 4);
        DRAWSEGMENT(1, 2);
    }
    else 
       DRAWSEGMENT(0, 1);
/*        newras = raster; */

    picture = mem_point(screenwidth, screenheight, 8, newras);
    pw_rop(pw, left, top, width, height,
                PIX_SRC, picture, 0, 0); 

    /*if (interleaved) */
        free(newras);
    panel_pw_text(pn, 250, 5+panel_fonthome(pf), PIX_SET, pf, "       ");
    return drawflag;

} /* rasterize */

#else

rasterize(interleaved, raster)
int interleaved;
register unsigned char *raster;
{
    register row, col;
    register c, lc;
   
fprintf(stderr, " you use the slow version\n");

#define DRAWSEGMENT(offset, step)                                        \
        for (row = top+offset; row < top+height; row += step) {          \
            notify_dispatch();                                           \
            if (!drawflag)                                               \
                goto quitdraw;                                           \
            for (col = left; col < left+width; col++)                    \
              if ((c = *raster++) != bgcolor || dobgflag)                \
                pw_rop(pw, size*col, size*row, size, size,               \
                           PIX_SRC | PIX_COLOR(c), (Pixrect *) 0, 0, 0); \
            if (lc++ == showcount) {                                     \
                lc = 0;                                                  \
                pw_show(pw);                                             \
            }                                                            \
        }


    panel_pw_text(pn, 250, 5+panel_fonthome(pf), PIX_SRC, pf, "Drawing");
    pw_batch_on(pw);
    drawflag = TRUE;
    lc = 0;
    if (interleaved) {
        DRAWSEGMENT(0, 8);
        DRAWSEGMENT(4, 8);
        DRAWSEGMENT(2, 4);
        DRAWSEGMENT(1, 2);
    }
    else {
        DRAWSEGMENT(0, 1);
    }

quitdraw:
    pw_batch_off(pw);
    panel_pw_text(pn, 250, 5+panel_fonthome(pf), PIX_SRC, pf, "       ");
    return drawflag;

} /* rasterize */

#endif


void waitevent()
{

    while(endpicflag == FALSE) {
        notify_dispatch();
        usleep(50000);
    }
    endpicflag = FALSE;

} /* waitevent */



void handlevent(canvas, event, arg)
Canvas *canvas;
Event *event;
char *arg;
{

    switch (event_action(event)) {
        case MS_RIGHT:
                if (event_is_down(event))
#ifdef oldSunOS
                exit(0);	/* You won't miss the following fancy stuff.. */
#else
                    if (alert_prompt(fr, (Event *)0, ALERT_MESSAGE_STRINGS,
		          "Please confirm:", "Do you know what you're doing??",
		          0,
		          ALERT_BUTTON_YES, "Of course, quit bugging me!",
		          ALERT_BUTTON_NO, "Sorry, I hit the wrong button...",
		          ALERT_MESSAGE_FONT, bf,
		          ALERT_BUTTON_FONT, pf,
		          0)
                          == ALERT_YES) 
                        exit(0);
                    else
                        notify_veto_destroy(fr);
#endif
                break;
        case MS_LEFT:
                if (event_is_down(event))
                    drawflag = FALSE;
                break;
        case MS_MIDDLE:
                if (event_is_down(event))
                    endpicflag = TRUE;
                break;
    }

} /* handlevent */



void usage()
{

        fprintf(stderr, "usage: %s [-l<num] [-s<num>] [-b] gif-files\n",
                        progname);

} /* usage */

